#!/usr/bin/env python

"""
View annotated sources.

Copyright (C) 2006 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from compiler.visitor import ASTVisitor
from simplified import *
import sys
import textwrap

# Exceptions.

class ViewerError(SimplifiedError):

    "An error in viewing."

    pass

# Classes.

class Viewer(ASTVisitor):

    """
    A viewing visitor for AST nodes.
    """

    def __init__(self, stream):
        ASTVisitor.__init__(self)
        self.cached_files = {}
        self.printed_lines = {}
        self.visitor = self
        self.stream = stream

    def process(self, module):
        self.dispatch(module)

    def dispatch(self, node):
        self.dispatch_only(node)
        ASTVisitor.dispatch(self, node)

    def dispatch_only(self, node, every_time=0):
        self.print_line(getattr(node, "filename", None), getattr(node, "lineno", None), every_time)

    def print_line(self, filename, lineno, every_time):
        last_printed = self.printed_lines.get(filename, 0)
        if lineno > last_printed or every_time:
            self.stream.write(self.get_line(filename, lineno))
            self.printed_lines[filename] = lineno

    def get_line(self, filename, lineno):
        if filename is None or lineno is None:
            return ""

        if self.cached_files.has_key(filename):
            lines = self.cached_files[filename]
        else:
            f = open(filename)
            try:
                self.cached_files[filename] = lines = f.readlines()
            finally:
                f.close()

        try:
            return lines[lineno - 1]
        except IndexError:
            return ""

    def report(self, exc):
        self.stream.write("Exception was:\n\n" + str(exc.exc) + "\n\n")
        self.stream.write("Nodes:\n\n")
        for node in exc.nodes:
            self.stream.write(repr(node) + "\n")
            self.dispatch_only(node.original, every_time=1)
        self.stream.write("\nOriginal node was:\n\n" + repr(exc.nodes[0].original) + "\n")
        self.stream.write("\nSimplified node was:\n\n")
        exc.nodes[0].pprint(stream=self.stream)

# HTML-related output production.

html_header = """<?xml version="1.0" encoding="iso-8859-15"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
  <title>Module</title>
  <style type="text/css">
    body {
      padding-top: 4em; padding-bottom: 4em;
      font-size: 14pt; font-family: monospace;
      background-color: black; color: white;
    }

    .class { margin-bottom: 1em; }
    .function { margin-bottom: 1em; }
    .body { padding-left: 2em; }
    .keyword { color: yellow; }
    .comment { color: blue; }
    .str { color: #FF00FF; }
    .doc { color: #FF00FF; margin-bottom: 1em; }
    .ref { color: cyan; }
    .ref a { color: cyan; text-decoration: none; }

    .popup {
      display: none; z-index: 2;
      position: absolute; top: 1em; left: 0.5em;
      padding: 0.5em; background-color: #000000;
    }

    .types {
      padding: 0.5em; background-color: #0000FF;
      float: right;
    }

    .scopes {
      padding: 0.5em; background-color: #007700;
      float: left;
    }

    .name,
    .attr
    {
      position: relative;
    }

    .name:hover > .popup,
    .attr:hover > .popup
    {
       display: block;
    }

  </style>
</head>
<body>
"""

html_footer = """</body>
</html>
"""

# Validation functions.

def hasnode(fn):
    def _hasnode(self, node):
        if not hasattr(node, "_node"):
            return
        else:
            fn(self, node)
    return _hasnode

# Browser classes.

class Browser(ASTVisitor):

    """
    A browsing visitor for AST nodes.
    """

    def __init__(self, stream):
        ASTVisitor.__init__(self)
        self.visitor = self
        self.stream = stream

    def process(self, module):
        self.stream.write(html_header)
        self.dispatch(module)
        self.stream.write(html_footer)

    def dispatch(self, node):
        try:
            ASTVisitor.dispatch(self, node)
        except ViewerError, exc:
            exc.add(node)
            raise
        except Exception, exc:
            raise ViewerError(exc, node)

    def visitModule(self, node):
        self.default(node)

    # Statements.

    def visitAssign(self, node):
        self.stream.write("<div class='assign'>\n")
        for lvalue in node.nodes:
            self.dispatch(lvalue)
            self.stream.write("=\n")
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitClass(self, node):
        definition = node._node
        structure = definition.expr.ref
        self.stream.write("<div class='class' id='%s'>\n" % self._url(structure.full_name()))
        self.stream.write("<p>\n")
        self._keyword("class")
        self._name_start(structure.name)
        self._popup_start()
        self._scopes(definition)
        self._popup_end()
        self._name_end()
        bases = structure.bases
        if bases:
            self.stream.write("(")
            first = 1
            for base in bases:
                if not first:
                    self.stream.write(",\n")
                self._name_start(base.name)
                self._popup_start()
                self._types(base)
                self._scopes(base)
                self._popup_end()
                self._name_end()
                first = 0
            self.stream.write(")")
        self.stream.write(":\n")
        self._comment(self._text(structure.full_name()))
        self.stream.write("</p>\n")

        self.stream.write("<div class='body'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    visitClass = hasnode(visitClass) # Remove unannotated nodes.

    def visitFunction(self, node):
        definition = node._node
        subprogram = definition.expr.ref
        self.stream.write("<div class='def' id='%s'>\n" % self._url(subprogram.full_name()))
        self.stream.write("<p>\n")
        self._keyword("def")
        self._name_start(subprogram.name)
        self._popup_start()
        self._scopes(definition)
        self._popup_end()
        self._name_end()
        self.stream.write("(")
        first = 1
        for param, default in subprogram.params:
            if not first:
                self.stream.write(",\n")
            self._parameter(subprogram, param)
            first = 0
        if subprogram.star is not None:
            if not first:
                self.stream.write(", *\n")
            param, default = subprogram.star
            self._parameter(subprogram, param)
            first = 0
        if subprogram.dstar is not None:
            if not first:
                self.stream.write(", **\n")
            param, default = subprogram.dstar
            self._parameter(subprogram, param)
            first = 0
        self.stream.write(")")
        self.stream.write(":\n")
        self._comment(self._text(subprogram.full_name()))
        self.stream.write("</p>\n")

        self.stream.write("<div class='body'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    visitFunction = hasnode(visitFunction) # Remove unannotated nodes.

    def visitPass(self, node):
        self._keyword("pass")

    def visitReturn(self, node):
        self._keyword("return")
        self.dispatch(node.value)

    def visitStmt(self, node):
        self.stream.write("<div class='stmt'>\n")
        self.default(node)
        self.stream.write("</div>\n")

    # Expressions.

    def visitCallFunc(self, node):
        self.stream.write("<span class='callfunc'>\n")
        self.dispatch(node.node)
        self.stream.write("(")
        first = 1
        for arg in node.args:
            if not first:
                self.stream.write(",\n")
            self.dispatch(arg)
            first = 0
        if node.star_args is not None:
            if not first:
                self.stream.write(", *\n")
            self.dispatch(node.star_args)
            first = 0
        if node.dstar_args is not None:
            if not first:
                self.stream.write(", **\n")
            self.dispatch(node.dstar_args)
            first = 0
        self.stream.write(")\n")
        self.stream.write("</span>\n")

    def visitTuple(self, node):
        self.stream.write("<span class='tuple'>\n")
        self.stream.write("(")
        self._sequence(node)
        self.stream.write(")\n")
        self.stream.write("</span>\n")

    visitAssTuple = visitTuple

    def visitList(self, node):
        self.stream.write("<span class='list'>\n")
        self.stream.write("[")
        self._sequence(node)
        self.stream.write("]\n")
        self.stream.write("</span>\n")

    visitAssList = visitList

    def visitName(self, node):
        if hasattr(node, "_node"):
            self._name_start(node._node.name)
            self._popup_start()
            self._types(node._node)
            self._scopes(node._node)
            self._popup_end()
            self._name_end()
        else:
            self._name(node.name)

    def visitAssName(self, node):
        if hasattr(node, "_node"):
            self._name_start(node._node.name)
            self._popup_start()
            self._types(node._node.expr)
            self._scopes(node._node)
            self._popup_end()
            self._name_end()
        else:
            self._name(node.name)

    def visitConst(self, node):
        self.stream.write(repr(node.value))

    def visitGetattr(self, node):
        self.stream.write("<span class='getattr'>\n")
        self.dispatch(node.expr)
        self.stream.write("<span class='attr'>\n")
        self.stream.write(".%s\n" % self._text(node.attrname))
        if hasattr(node, "_node"):
            self._popup_start()
            self._types(node._node)
            self._scopes(node._node)
            self._popup_end()
        self.stream.write("</span>\n")
        self.stream.write("</span>\n")

    def visitAssAttr(self, node):
        self.stream.write("<span class='assattr'>\n")
        self.dispatch(node.expr)
        self.stream.write("<span class='attr'>\n")
        self.stream.write(".%s\n" % self._text(node.attrname))
        if hasattr(node, "_node"):
            self._popup_start()
            self._types(node._node)
            self._scopes(node._node)
            self._popup_end()
        self.stream.write("</span>\n")
        self.stream.write("</span>\n")

    # Output preparation methods.

    def _text(self, text):
        return text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

    def _attr(self, attr):
        return self._text(attr).replace("'", "&apos;").replace('"', "&quot;")

    def _url(self, url):
        return self._attr(url).replace("#", "%23").replace("-", "%2d")

    def _comment(self, comment):
        self.stream.write("<span class='comment'># %s</span>\n" % comment)

    def _keyword(self, kw):
        self.stream.write("<span class='keyword'>%s</span> " % kw)

    def _doc(self, node):
        if node.doc is not None:
            self.stream.write("<pre class='doc'>\n")
            self.stream.write('"""')
            output = textwrap.dedent(node.doc.replace('"""', '\\"\\"\\"'))
            self.stream.write(self._text(output))
            self.stream.write('"""')
            self.stream.write("</pre>\n")

    def _sequence(self, node):
        first = 1
        for n in node.nodes:
            if not first:
                self.stream.write(",\n")
            self.dispatch(n)
            first = 0

    def _parameter(self, subprogram, param):
        self._name_start(param)
        if hasattr(subprogram, "paramtypes"):
            self._popup_start()
            self._types_list(subprogram.paramtypes[param])
            self._popup_end()
        self._name_end()

    def _name(self, name):
        self.stream.write("<span class='name'>%s</span>\n" % name)

    def _name_start(self, name):
        self.stream.write("<span class='name'>%s\n" % name)

    def _name_end(self):
        self.stream.write("</span>\n")

    def _popup_start(self):
        self.stream.write("<span class='popup'>\n")

    def _popup_end(self):
        self.stream.write("</span>\n")

    def _types(self, node):
        if hasattr(node, "types"):
            self._types_list(node.types)
        elif hasattr(node, "writes"):
            self._types_list(flatten(node.writes.values()))
        elif hasattr(node, "accesses"):
            self._types_list(flatten(node.accesses.values()))
        else:
            self.stream.write("<div class='types'>\n")
            self.stream.write("No types!\n")
            self.stream.write("</div>\n")

    def _types_list(self, types):
        self.stream.write("<div class='types'>\n")
        for type in types:
            fn = type.type.full_name()
            self.stream.write("<div class='type'>")
            self.stream.write(self._text(fn)) 
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def _scopes(self, node):
        if not isinstance(node, LoadName):
            self.stream.write("<div class='scopes'>\n")
            if not hasattr(node, "writes") and not hasattr(node, "accesses"):
                self.stream.write("No scopes!\n")
            else:
                for ref in getattr(node, "writes", getattr(node, "accesses", {})).keys():
                    fn = ref.full_name()
                    self.stream.write("<div class='scope'>")
                    self.stream.write(self._text(fn)) 
                    self.stream.write("</div>\n")
            self.stream.write("</div>\n")

# Utility functions.

def flatten(lists):
    result = []
    for l in lists:
        for attr in l:
            if attr not in result:
                result.append(attr)
    return result

# Convenience functions.

def view(module, stream=None):
    viewer = Viewer(stream or sys.stdout)
    viewer.process(module.original)

def browse(module, stream=None):
    browser = Browser(stream or sys.stdout)
    browser.process(module.original)

def makedoc(module, filename):
    stream = open(filename, "wb")
    try:
        browser = Browser(stream)
        browser.process(module.original)
    finally:
        stream.close()

# vim: tabstop=4 expandtab shiftwidth=4
